#ifndef TEST_COMMON

#define TEST_COMMON 1

#ifndef _GNU_SOURCE
  #define _GNU_SOURCE
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <check.h>
#include "../src/slapt.h"

#endif
